@@ -26,21 +26,28 @@
 
 #include "pygobject-private.h"
 #include "pythread.h"
+#include <gobject/gvaluecollector.h>
 
 static PyObject *gerror_exc = NULL;
-static const gchar *pyginterface_type_id   = "PyGInterface::type";
-GQuark pyginterface_type_key  = 0;
-static const gchar *pygobject_class_init_id   = "PyGObject::class-init";
-GQuark pygobject_class_init_key  = 0;
-static const gchar *pyginterface_info_id   = "PyGInterface::info";
-GQuark pyginterface_info_key  = 0;
+static gboolean use_gil_state_api = FALSE;
+
+GQuark pyginterface_type_key;
+GQuark pygobject_class_init_key;
+GQuark pyginterface_info_key;
+GQuark pygboxed_type_key;
+GQuark pygobject_class_key;
+GQuark pygobject_wrapper_key;
+GQuark pygboxed_marshal_key;
+GQuark pygenum_class_key;
+GQuark pygflags_class_key;
+GQuark pygpointer_class_key;
+GQuark pygobject_has_updated_constructor_key;
+
 
 
 static void pyg_flags_add_constants(PyObject *module, GType flags_type,
 				    const gchar *strip_prefix);
-static gboolean pyg_error_check(GError **error);
 
-static gboolean use_gil_state_api = FALSE;
     
 /* -------------- GDK threading hooks ---------------------------- */
 
@@ -235,6 +242,11 @@ pyg_type_name (PyObject *self, PyObject *args)
     GType type;
     const gchar *name;
 
+    if (PyErr_Warn(PyExc_DeprecationWarning,
+		   "gobject.type_name is deprecated; "
+		   "use GType.name instead"))
+        return NULL;
+    
     if (!PyArg_ParseTuple(args, "O:gobject.type_name", &gtype))
 	return NULL;
     if ((type = pyg_type_from_object(gtype)) == 0)
@@ -252,6 +264,11 @@ pyg_type_from_name (PyObject *self, PyObject *args)
     const gchar *name;
     GType type;
 
+    if (PyErr_Warn(PyExc_DeprecationWarning,
+		   "gobject.type_from_name is deprecated; "
+		   "use GType.from_name instead"))
+        return NULL;
+    
     if (!PyArg_ParseTuple(args, "s:gobject.type_from_name", &name))
 	return NULL;
     type = g_type_from_name(name);
@@ -267,6 +284,11 @@ pyg_type_parent (PyObject *self, PyObject *args)
     PyObject *gtype;
     GType type, parent;
 
+    if (PyErr_Warn(PyExc_DeprecationWarning,
+		   "gobject.type_parent is deprecated; "
+		   "use GType.parent instead"))
+        return NULL;
+    
     if (!PyArg_ParseTuple(args, "O:gobject.type_parent", &gtype))
 	return NULL;
     if ((type = pyg_type_from_object(gtype)) == 0)
@@ -284,13 +306,18 @@ pyg_type_is_a (PyObject *self, PyObject *args)
     PyObject *gtype, *gparent;
     GType type, parent;
 
+    if (PyErr_Warn(PyExc_DeprecationWarning,
+		   "gobject.type_is_a is deprecated; "
+		   "use GType.is_a instead"))
+        return NULL;
+    
     if (!PyArg_ParseTuple(args, "OO:gobject.type_is_a", &gtype, &gparent))
 	return NULL;
     if ((type = pyg_type_from_object(gtype)) == 0)
 	return NULL;
     if ((parent = pyg_type_from_object(gparent)) == 0)
 	return NULL;
-    return PyInt_FromLong(g_type_is_a(type, parent));
+    return PyBool_FromLong(g_type_is_a(type, parent));
 }
 
 static PyObject *
@@ -300,6 +327,11 @@ pyg_type_children (PyObject *self, PyObject *args)
     GType type, *children;
     guint n_children, i;
 
+    if (PyErr_Warn(PyExc_DeprecationWarning,
+		   "gobject.type_children is deprecated; "
+		   "use GType.children instead"))
+        return NULL;
+    
     if (!PyArg_ParseTuple(args, "O:gobject.type_children", &gtype))
 	return NULL;
     if ((type = pyg_type_from_object(gtype)) == 0)
@@ -326,6 +358,11 @@ pyg_type_interfaces (PyObject *self, PyObject *args)
     GType type, *interfaces;
     guint n_interfaces, i;
 
+    if (PyErr_Warn(PyExc_DeprecationWarning,
+		   "gobject.type_interfaces is deprecated; "
+		   "use GType.interfaces instead"))
+        return NULL;
+    
     if (!PyArg_ParseTuple(args, "O:gobject.type_interfaces", &gtype))
 	return NULL;
     if ((type = pyg_type_from_object(gtype)) == 0)
@@ -497,17 +534,19 @@ override_signal(GType instance_type, const gchar *signal_name)
     return TRUE;
 }
 
-static gboolean
+static PyObject *
 add_signals (GType instance_type, PyObject *signals)
 {
     gboolean ret = TRUE;
     GObjectClass *oclass;
     int pos = 0;
-    PyObject *key, *value;
+    PyObject *key, *value, *overridden_signals = NULL;
 
+    overridden_signals = PyDict_New();
     oclass = g_type_class_ref(instance_type);
     while (PyDict_Next(signals, &pos, &key, &value)) {
 	const gchar *signal_name;
+        gchar *signal_name_canon, *c;
 
 	if (!PyString_Check(key)) {
 	    PyErr_SetString(PyExc_TypeError,
@@ -519,7 +558,20 @@ add_signals (GType instance_type, PyObject *signals)
 
 	if (value == Py_None ||
 	    (PyString_Check(value) &&
-	     !strcmp(PyString_AsString(value), "override"))) {
+	     !strcmp(PyString_AsString(value), "override")))
+        {
+              /* canonicalize signal name, replacing '-' with '_' */
+            signal_name_canon = g_strdup(signal_name);
+            for (c = signal_name_canon; *c; ++c)
+                if (*c == '-')
+                    *c = '_';
+            if (PyDict_SetItemString(overridden_signals, signal_name_canon, key)) {
+                g_free(signal_name_canon);
+                ret = FALSE;
+                break;
+            }
+            g_free(signal_name_canon);
+
 	    ret = override_signal(instance_type, signal_name);
 	} else {
 	    ret = create_signal(instance_type, signal_name, value);
@@ -529,7 +581,12 @@ add_signals (GType instance_type, PyObject *signals)
 	    break;
     }
     g_type_class_unref(oclass);
-    return ret;
+    if (ret)
+        return overridden_signals;
+    else {
+        Py_XDECREF(overridden_signals);
+        return NULL;
+    }
 }
 
 static GParamSpec *
@@ -868,12 +925,17 @@ add_properties (GType instance_type, PyObject *properties)
 static void
 pyg_register_class_init(GType gtype, PyGClassInitFunc class_init)
 {
-    g_type_set_qdata(gtype, pygobject_class_init_key, class_init);
+    GSList *list;
+
+    list = g_type_get_qdata(gtype, pygobject_class_init_key);
+    list = g_slist_prepend(list, class_init);
+    g_type_set_qdata(gtype, pygobject_class_init_key, list);
 }
 
 static int
 pyg_run_class_init(GType gtype, gpointer gclass, PyTypeObject *pyclass)
 {
+    GSList *list;
     PyGClassInitFunc class_init;
     GType parent_type;
     int rv;
@@ -881,25 +943,154 @@ pyg_run_class_init(GType gtype, gpointer gclass, PyTypeObject *pyclass)
     parent_type = g_type_parent(gtype);
     if (parent_type) {
         rv = pyg_run_class_init(parent_type, gclass, pyclass);
-        if (rv) return rv;
+        if (rv)
+	    return rv;
     }
-    class_init = g_type_get_qdata(gtype, pygobject_class_init_key);
-    if (class_init)
-        return class_init(gclass, pyclass);
+    
+    list = g_type_get_qdata(gtype, pygobject_class_init_key);
+    for (; list; list = list->next) {
+	class_init = list->data;
+        rv = class_init(gclass, pyclass);
+        if (rv)
+	    return rv;
+    }
+    
     return 0;
 }
 
-
 static PyObject *
-pyg_type_register(PyObject *self, PyObject *args)
+_wrap_pyg_type_register(PyObject *self, PyObject *args)
 {
-    PyObject *gtype, *module, *gsignals, *gproperties;
     PyTypeObject *class;
-    GType parent_type, instance_type;
-    gchar *type_name = NULL;
+    PyObject *gtype;
+
+    if (PyErr_Warn(PyExc_DeprecationWarning, "gobject.type_register is deprecated;"
+                   " now types are automatically registered"))
+        return NULL;
+
+    if (!PyArg_ParseTuple(args, "O:gobject.type_register", &class))
+	return NULL;
+    if (!PyType_IsSubtype(class, &PyGObject_Type)) {
+	PyErr_SetString(PyExc_TypeError,"argument must be a GObject subclass");
+	return NULL;
+    }
+
+      /* Check if type already registered */
+    gtype = PyObject_GetAttrString((PyObject *)class, "__gtype__");
+    if (gtype == NULL) {
+        PyErr_Clear();
+          /* not registered */
+        if (pyg_type_register(class, NULL))
+            return NULL;
+    } else
+          /* already registered */
+        Py_DECREF(gtype);
+
+    Py_INCREF(class);
+    return (PyObject *) class;
+}
+
+static char *
+get_type_name_for_class(PyTypeObject *class)
+{
     gint i, name_serial;
+    char name_serial_str[16];
+    PyObject *module;
+    char *type_name;
+    
+    /* make name for new GType */
+    name_serial = 1;
+    /* give up after 1000 tries, just in case.. */
+    while (name_serial < 1000) 
+    {
+	snprintf(name_serial_str, 16, "-v%i", name_serial);
+	module = PyObject_GetAttrString((PyObject *)class, "__module__");
+	if (module && PyString_Check(module)) {
+	    type_name = g_strconcat(PyString_AsString(module), ".",
+				    class->tp_name,
+				    name_serial > 1 ? name_serial_str : NULL,
+				    NULL);
+	    Py_DECREF(module);
+	} else {
+	    if (module)
+		Py_DECREF(module);
+	    else
+		PyErr_Clear();
+	    type_name = g_strconcat(class->tp_name,
+				    name_serial > 1 ? name_serial_str : NULL,
+				    NULL);
+	}
+	/* convert '.' in type name to '+', which isn't banned (grumble) */
+	for (i = 0; type_name[i] != '\0'; i++)
+	    if (type_name[i] == '.')
+		type_name[i] = '+';
+	if (g_type_from_name(type_name) == 0)
+	    break;              /* we now have a unique name */
+	++name_serial;
+    }
+
+    return type_name;
+}
+
+
+static GStaticPrivate pygobject_contruction_wrapper = G_STATIC_PRIVATE_INIT;
+
+static inline void
+pygobject_init_wrapper_set(PyObject *wrapper)
+{
+    g_static_private_set(&pygobject_contruction_wrapper, wrapper, NULL);
+}
+
+static inline PyObject *
+pygobject_init_wrapper_get(void)
+{
+    return (PyObject *) g_static_private_get(&pygobject_contruction_wrapper);
+}
+
+static void
+pygobject__g_instance_init(GTypeInstance   *instance,
+                           gpointer         g_class)
+{
+    GObject *object = (GObject *) instance;
+    PyObject *wrapper, *args, *kwargs;
+
+    if (!g_type_get_qdata(G_OBJECT_TYPE(object), pygobject_has_updated_constructor_key))
+        return;
+
+    wrapper = g_object_get_qdata(object, pygobject_wrapper_key); 
+    if (wrapper == NULL) {
+        wrapper = pygobject_init_wrapper_get();
+        if (wrapper) {
+            g_object_set_qdata(object, pygobject_wrapper_key, wrapper);
+        }
+    }
+    pygobject_init_wrapper_set(NULL);
+    if (wrapper == NULL) {
+          /* this looks like a python object created through
+           * g_object_new -> we have no python wrapper, so create it
+           * now */
+        PyGILState_STATE state;
+        state = pyg_gil_state_ensure();
+        wrapper = pygobject_new_full(object, FALSE);
+        args = PyTuple_New(0);
+        kwargs = PyDict_New();
+        if (wrapper->ob_type->tp_init(wrapper, args, kwargs))
+            PyErr_Print();
+        Py_DECREF(args);
+        Py_DECREF(kwargs);
+        pyg_gil_state_release(state);
+    }
+}
+
+int
+pyg_type_register(PyTypeObject *class, char *type_name)
+{
+    PyObject *gtype, *gsignals, *gproperties, *overridden_signals;
+    GType parent_type, instance_type;
+    gint i;
     GTypeQuery query;
     gpointer gclass;
+    gpointer has_new_constructor_api;
     GTypeInfo type_info = {
 	0,    /* class_size */
 
@@ -912,54 +1103,19 @@ pyg_type_register(PyObject *self, PyObject *args)
 
 	0,    /* instance_size */
 	0,    /* n_preallocs */
-	(GInstanceInitFunc) NULL
+	(GInstanceInitFunc) pygobject__g_instance_init
     };
 
-    if (!PyArg_ParseTuple(args, "O:gobject.type_register", &class))
-	return NULL;
-    if (!PyType_IsSubtype(class, &PyGObject_Type)) {
-	PyErr_SetString(PyExc_TypeError,"argument must be a GObject subclass");
-	return NULL;
-    }
 
     /* find the GType of the parent */
     parent_type = pyg_type_from_object((PyObject *)class);
     if (!parent_type) {
-	return NULL;
-    }
-
-      /* make name for new GType */
-    name_serial = 1;
-    while (name_serial < 1000) /* give up after 1000 tries, just in case.. */
-    {
-        char name_serial_str[16];
-
-        snprintf(name_serial_str, 16, "-v%i", name_serial);
-        module = PyObject_GetAttrString((PyObject *)class, "__module__");
-        if (module && PyString_Check(module)) {
-            type_name = g_strconcat(PyString_AsString(module), ".",
-                                    class->tp_name,
-                                    name_serial > 1? name_serial_str : NULL,
-                                    NULL);
-	    Py_DECREF(module);
-        } else {
-            if (module)
-                Py_DECREF(module);
-            else
-                PyErr_Clear();
-            type_name = g_strconcat(class->tp_name,
-                                    name_serial > 1? name_serial_str : NULL,
-                                    NULL);
-        }
-          /* convert '.' in type name to '+', which isn't banned (grumble) */
-        for (i = 0; type_name[i] != '\0'; i++)
-            if (type_name[i] == '.')
-                type_name[i] = '+';
-        if (g_type_from_name(type_name) == 0)
-            break;              /* we now have a unique name */
-        ++name_serial;
+	return -1;
     }
 
+    if (!type_name)
+	type_name = get_type_name_for_class(class);
+    
     /* set class_data that will be passed to the class_init function. */
     type_info.class_data = class;
 
@@ -974,7 +1130,7 @@ pyg_type_register(PyObject *self, PyObject *args)
     g_free(type_name);
     if (instance_type == 0) {
 	PyErr_SetString(PyExc_RuntimeError, "could not create new GType");
-	return NULL;
+	return -1;
     }
 
     /* store pointer to the class with the GType */
@@ -987,6 +1143,16 @@ pyg_type_register(PyObject *self, PyObject *args)
     PyDict_SetItemString(class->tp_dict, "__gtype__", gtype);
     Py_DECREF(gtype);
 
+      /* propagate new constructor API compatility flag from parent to child type */
+    has_new_constructor_api = g_type_get_qdata(parent_type,
+                                               pygobject_has_updated_constructor_key);
+    if (has_new_constructor_api == NULL)
+        g_warning("Constructor wrapper for %s needs to be updated to the new API",
+                  g_type_name(parent_type));
+    else
+        g_type_set_qdata(instance_type, pygobject_has_updated_constructor_key,
+                         has_new_constructor_api);
+
     /* if no __doc__, set it to the auto doc descriptor */
     if (PyDict_GetItemString(class->tp_dict, "__doc__") == NULL) {
 	PyDict_SetItemString(class->tp_dict, "__doc__",
@@ -1000,13 +1166,14 @@ pyg_type_register(PyObject *self, PyObject *args)
 	if (!PyDict_Check(gsignals)) {
 	    PyErr_SetString(PyExc_TypeError,
 			    "__gsignals__ attribute not a dict!");
-	    return NULL;
+	    return -1;
 	}
-	if (!add_signals(instance_type, gsignals)) {
-	    return NULL;
+	if (!(overridden_signals = add_signals(instance_type, gsignals))) {
+	    return -1;
 	}
-	PyDict_DelItemString(class->tp_dict, "__gsignals__");
-	/* Borrowed reference. Py_DECREF(gsignals); */
+        if (PyDict_SetItemString(class->tp_dict, "__gsignals__", overridden_signals))
+            return -1;
+        Py_DECREF(overridden_signals);
     } else {
 	PyErr_Clear();
     }
@@ -1018,10 +1185,10 @@ pyg_type_register(PyObject *self, PyObject *args)
 	if (!PyDict_Check(gproperties)) {
 	    PyErr_SetString(PyExc_TypeError,
 			    "__gproperties__ attribute not a dict!");
-	    return NULL;
+	    return -1;
 	}
 	if (!add_properties(instance_type, gproperties)) {
-	    return NULL;
+	    return -1;
 	}
 	PyDict_DelItemString(class->tp_dict, "__gproperties__");
 	/* Borrowed reference. Py_DECREF(gproperties); */
@@ -1032,10 +1199,13 @@ pyg_type_register(PyObject *self, PyObject *args)
     gclass = g_type_class_ref(instance_type);
     if (pyg_run_class_init(instance_type, gclass, class)) {
         g_type_class_unref(gclass);
-        return NULL;
+        return -1;
     }
     g_type_class_unref(gclass);
 
+    if (gsignals)
+        PyDict_DelItemString(class->tp_dict, "__gsignals__");
+
       /* Register interface implementations  */
     if (class->tp_bases) {
         for (i = 0; i < PyTuple_GET_SIZE(class->tp_bases); ++i)
@@ -1063,8 +1233,7 @@ pyg_type_register(PyObject *self, PyObject *args)
     } else
         g_warning("type has no tp_bases");
 
-    Py_INCREF(class);
-    return (PyObject *) class;
+    return 0;
 }
 
 static PyObject *
@@ -1443,7 +1612,8 @@ pyg_object_new (PyGObject *self, PyObject *args, PyObject *kwargs)
 	    }
 	    g_value_init(&params[n_params].value,
 			 G_PARAM_SPEC_VALUE_TYPE(pspec));
-	    if (pyg_value_from_pyobject(&params[n_params].value, value)) {
+	    if (pyg_param_gvalue_from_pyobject(&params[n_params].value,
+					       value, pspec) < 0) {
 		PyErr_Format(PyExc_TypeError,
 			     "could not convert value for property `%s'",
 			     key_str);
@@ -1467,8 +1637,12 @@ pyg_object_new (PyGObject *self, PyObject *args, PyObject *kwargs)
     g_type_class_unref(class);
     
     if (obj)
-	return pygobject_new ((GObject *)obj);
-    return NULL;
+	self = (PyGObject *) pygobject_new ((GObject *)obj);
+    else
+        self = NULL;
+    g_object_unref(obj);
+
+    return (PyObject *) self;
 }
 
 static gint
@@ -1511,8 +1685,8 @@ get_handler_priority(gint *priority, PyObject *kwargs)
     return 0;
 }
 
-static gboolean
-handler_marshal(gpointer user_data)
+gboolean
+pyg_handler_marshal(gpointer user_data)
 {
     PyObject *tuple, *ret;
     gboolean res;
@@ -1566,13 +1740,13 @@ pyg_idle_add(PyObject *self, PyObject *args, PyObject *kwargs)
 
     cbargs = PySequence_GetSlice(args, 1, len);
     if (cbargs == NULL)
-      return NULL;
+	return NULL;
 
     data = Py_BuildValue("(ON)", callback, cbargs);
     if (data == NULL)
-      return NULL;
-    handler_id = g_idle_add_full(priority, handler_marshal, data,
-				 (GDestroyNotify)pyg_destroy_notify);
+	return NULL;
+    handler_id = g_idle_add_full(priority, pyg_handler_marshal, data,
+				 pyg_destroy_notify);
     return PyInt_FromLong(handler_id);
 }
 
@@ -1580,8 +1754,8 @@ static PyObject *
 pyg_timeout_add(PyObject *self, PyObject *args, PyObject *kwargs)
 {
     PyObject *first, *callback, *cbargs = NULL, *data;
-    gint len, priority = G_PRIORITY_DEFAULT, interval;
-    guint handler_id;
+    gint len, priority = G_PRIORITY_DEFAULT;
+    guint interval, handler_id;
 
     len = PyTuple_Size(args);
     if (len < 2) {
@@ -1590,7 +1764,7 @@ pyg_timeout_add(PyObject *self, PyObject *args, PyObject *kwargs)
 	return NULL;
     }
     first = PySequence_GetSlice(args, 0, 2);
-    if (!PyArg_ParseTuple(first, "iO:timeout_add", &interval, &callback)) {
+    if (!PyArg_ParseTuple(first, "IO:timeout_add", &interval, &callback)) {
 	Py_DECREF(first);
         return NULL;
     }
@@ -1604,13 +1778,14 @@ pyg_timeout_add(PyObject *self, PyObject *args, PyObject *kwargs)
 
     cbargs = PySequence_GetSlice(args, 2, len);
     if (cbargs == NULL)
-      return NULL;
+	return NULL;
 
     data = Py_BuildValue("(ON)", callback, cbargs);
     if (data == NULL)
-      return NULL;
-    handler_id = g_timeout_add_full(priority, interval, handler_marshal, data,
-				    (GDestroyNotify)pyg_destroy_notify);
+	return NULL;
+    handler_id = g_timeout_add_full(priority, interval,
+				    pyg_handler_marshal, data,
+				    pyg_destroy_notify);
     return PyInt_FromLong(handler_id);
 }
 
@@ -2048,6 +2223,42 @@ pyg_spawn_async(PyObject *unused, PyObject *args, PyObject *kwargs)
     return Py_BuildValue("iNNN", child_pid, pystdin, pystdout, pystderr);
 }
 
+
+static PyObject *
+pyg_markup_escape_text(PyObject *unused, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "text", NULL };
+    char *text_in, *text_out;
+    int text_size;
+    PyObject *retval;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s#:gobject.markup_escape_text", kwlist,
+                                     &text_in, &text_size))
+        return NULL;
+
+    text_out = g_markup_escape_text(text_in, text_size);
+    retval = PyString_FromString(text_out);
+    g_free(text_out);
+    return retval;
+}
+
+static PyObject *
+pyg_get_current_time(PyObject *unused)
+{
+    GTimeVal timeval;
+    double ret;
+
+    g_get_current_time(&timeval);
+    ret = (double)timeval.tv_sec + (double)timeval.tv_usec * 0.000001;
+    return PyFloat_FromDouble(ret);
+}
+
+static PyObject *
+pyg_main_depth(PyObject *unused)
+{
+    return PyInt_FromLong(g_main_depth());
+}
+
 static PyMethodDef pygobject_functions[] = {
     { "type_name", pyg_type_name, METH_VARARGS },
     { "type_from_name", pyg_type_from_name, METH_VARARGS },
@@ -2055,7 +2266,7 @@ static PyMethodDef pygobject_functions[] = {
     { "type_is_a", pyg_type_is_a, METH_VARARGS },
     { "type_children", pyg_type_children, METH_VARARGS },
     { "type_interfaces", pyg_type_interfaces, METH_VARARGS },
-    { "type_register", pyg_type_register, METH_VARARGS },
+    { "type_register", _wrap_pyg_type_register, METH_VARARGS },
     { "signal_new", pyg_signal_new, METH_VARARGS },
     { "signal_list_names", (PyCFunction)pyg_signal_list_names, METH_VARARGS|METH_KEYWORDS },
     { "signal_list_ids", (PyCFunction)pyg_signal_list_ids, METH_VARARGS|METH_KEYWORDS },
@@ -2072,6 +2283,9 @@ static PyMethodDef pygobject_functions[] = {
     { "threads_init", (PyCFunction)pyg_threads_init, METH_VARARGS|METH_KEYWORDS },
     { "child_watch_add", (PyCFunction)pyg_child_watch_add, METH_VARARGS|METH_KEYWORDS },
     { "spawn_async", (PyCFunction)pyg_spawn_async, METH_VARARGS|METH_KEYWORDS },
+    { "markup_escape_text", (PyCFunction)pyg_markup_escape_text, METH_VARARGS|METH_KEYWORDS },
+    { "get_current_time", (PyCFunction)pyg_get_current_time, METH_NOARGS },
+    { "main_depth", (PyCFunction)pyg_main_depth, METH_NOARGS },
 
     { NULL, NULL, 0 }
 };
@@ -2198,7 +2412,7 @@ pyg_flags_add_constants(PyObject *module, GType flags_type,
  *
  * Returns: True if an error was set.
  */
-static gboolean
+gboolean
 pyg_error_check(GError **error)
 {
     PyGILState_STATE state;
@@ -2330,6 +2544,97 @@ pyg_parse_constructor_args(GType        obj_type,
     return TRUE;
 }
 
+int
+pygobject_constructv(PyGObject  *self,
+                     guint       n_parameters,
+                     GParameter *parameters)
+{
+    if (self->obj == NULL) {
+        pygobject_init_wrapper_set((PyObject *) self);
+        self->obj = g_object_newv(pyg_type_from_object((PyObject *) self),
+                                  n_parameters, parameters);
+        pygobject_init_wrapper_set(NULL);
+        pygobject_register_wrapper((PyObject *) self);
+    } else {
+        int i;
+        for (i = 0; i < n_parameters; ++i)
+            g_object_set_property(self->obj,
+				  parameters[i].name,
+				  &parameters[i].value);
+    }
+    return 0;
+}
+
+/* This function is mostly juste copy-paste from g_object_new, but
+ * calls pygobject_constructv instead of g_object_newv */
+int
+pygobject_construct(PyGObject *self, const char *first_property_name, ...)
+{
+    va_list var_args;
+    GObjectClass *class;
+    GParameter *params;
+    const gchar *name;
+    guint n_params = 0, n_alloced_params = 16;
+    GType object_type = pyg_type_from_object((PyObject *) self);
+    int retval;
+
+    if (!first_property_name)
+        return pygobject_constructv(self, 0, NULL);
+
+    va_start(var_args, first_property_name);
+    class = g_type_class_ref(object_type);
+
+    params = g_new(GParameter, n_alloced_params);
+    name = first_property_name;
+    while (name)
+    {
+        gchar *error = NULL;
+        GParamSpec *pspec = g_object_class_find_property(class, name);
+
+        if (!pspec)
+	{
+            g_warning("%s: object class `%s' has no property named `%s'", 
+                      G_STRFUNC, 
+                      g_type_name(object_type), 
+                      name);
+            break;
+	}
+        if (n_params >= n_alloced_params)
+	{
+            n_alloced_params += 16;
+            params = g_renew(GParameter, params, n_alloced_params);
+	}
+        params[n_params].name = name;
+        params[n_params].value.g_type = 0;
+        g_value_init(&params[n_params].value, G_PARAM_SPEC_VALUE_TYPE(pspec));
+        G_VALUE_COLLECT(&params[n_params].value, var_args, 0, &error);
+        if (error)
+	{
+            g_warning("%s: %s", G_STRFUNC, error);
+            g_free(error);
+            g_value_unset(&params[n_params].value);
+            break;
+	}
+        n_params++;
+        name = va_arg(var_args, gchar*);
+    }
+
+    retval = pygobject_constructv(self, n_params, params);
+
+    while (n_params--)
+        g_value_unset(&params[n_params].value);
+    g_free(params);
+    va_end(var_args);
+    g_type_class_unref(class);
+    return retval;
+}
+
+void
+pyg_set_object_has_new_constructor(GType type)
+{
+    g_type_set_qdata(type, pygobject_has_updated_constructor_key, GINT_TO_POINTER(1));
+}
+
 /* ----------------- gobject module initialisation -------------- */
 
 struct _PyGObject_Functions pygobject_api_functions = {
@@ -2347,7 +2652,7 @@ struct _PyGObject_Functions pygobject_api_functions = {
   pyg_type_wrapper_new,
   pyg_enum_get_value,
   pyg_flags_get_value,
-  pyg_register_boxed_custom,
+  pyg_register_gtype_custom,
   pyg_value_from_pyobject,
   pyg_value_as_pyobject,
 
@@ -2394,7 +2699,12 @@ struct _PyGObject_Functions pygobject_api_functions = {
   pyg_gil_state_ensure_py23,
   pyg_gil_state_release_py23,
   pyg_register_class_init,
-  pyg_register_interface_info
+  pyg_register_interface_info,
+
+  pyg_closure_set_exception_handler,
+  pygobject_constructv,
+  pygobject_construct,
+  pyg_set_object_has_new_constructor
 };
 
 #define REGISTER_TYPE(d, type, name) \
@@ -2415,8 +2725,7 @@ DL_EXPORT(void)
 initgobject(void)
 {
     PyObject *m, *d, *o, *tuple;
-
-    PyGTypeWrapper_Type.ob_type = &PyType_Type;
+    PyObject *descr;
     PyGParamSpec_Type.ob_type = &PyType_Type;
     
     m = Py_InitModule("gobject", pygobject_functions);
@@ -2424,29 +2733,57 @@ initgobject(void)
 
     g_type_init();
 
+    pygboxed_type_key        = g_quark_from_static_string("PyGBoxed::class");
+    pygboxed_marshal_key     = g_quark_from_static_string("PyGBoxed::marshal");
+    pygenum_class_key        = g_quark_from_static_string("PyGEnum::class");
+    pygflags_class_key       = g_quark_from_static_string("PyGFlags::class");
+    pygobject_class_key      = g_quark_from_static_string("PyGObject::class");
+    pygobject_class_init_key = g_quark_from_static_string("PyGObject::class-init");
+    pygobject_wrapper_key    = g_quark_from_static_string("PyGObject::wrapper");
+    pyginterface_type_key    = g_quark_from_static_string("PyGInterface::type");
+    pyginterface_info_key    = g_quark_from_static_string("PyGInterface::info");
+    pygpointer_class_key     = g_quark_from_static_string("PyGPointer::class");
+    pygobject_has_updated_constructor_key =\
+        g_quark_from_static_string("PyGObject::has-updated-constructor");
+
+    REGISTER_TYPE(d, PyGTypeWrapper_Type, "GType");
+
     PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
 						  pyobject_copy,
 						  pyobject_free);
 
     gerror_exc = PyErr_NewException("gobject.GError", PyExc_RuntimeError,NULL);
     PyDict_SetItemString(d, "GError", gerror_exc);
-    
+
+    PyGObject_MetaType.tp_traverse = PyType_Type.tp_traverse;
+    PyGObject_MetaType.tp_clear = PyType_Type.tp_clear;
+    PyGObject_MetaType.tp_is_gc = PyType_Type.tp_is_gc;
+    PyType_Ready(&PyGObject_MetaType);
+    PyDict_SetItemString(d, "GObjectMeta", (PyObject *) &PyGObject_MetaType);
+
     PyGObject_Type.tp_alloc = PyType_GenericAlloc;
     PyGObject_Type.tp_new = PyType_GenericNew;
     pygobject_register_class(d, "GObject", G_TYPE_OBJECT,
 			     &PyGObject_Type, NULL);
     PyDict_SetItemString(PyGObject_Type.tp_dict, "__gdoc__",
 			 pyg_object_descr_doc_get());
+    pyg_set_object_has_new_constructor(G_TYPE_OBJECT);
+
+      /* GObject properties descriptor */
+    if (PyType_Ready(&PyGProps_Type) < 0)
+        return;
+    if (PyType_Ready(&PyGPropsDescr_Type) < 0)
+        return;
+    if (PyType_Ready(&PyGPropsIter_Type) < 0)
+        return;
+    descr = PyObject_New(PyObject, &PyGPropsDescr_Type);
+    PyDict_SetItemString(PyGObject_Type.tp_dict, "props", descr);
 
     REGISTER_GTYPE(d, PyGInterface_Type, "GInterface", G_TYPE_INTERFACE);
     PyDict_SetItemString(PyGInterface_Type.tp_dict, "__doc__",
 			 pyg_object_descr_doc_get());
     PyDict_SetItemString(PyGInterface_Type.tp_dict, "__gdoc__",
 			 pyg_object_descr_doc_get());
-    pyginterface_type_key = g_quark_from_static_string(pyginterface_type_id);
-    pyginterface_info_key = g_quark_from_static_string(pyginterface_info_id);
-
-    pygobject_class_init_key = g_quark_from_static_string(pygobject_class_init_id);
 
     REGISTER_GTYPE(d, PyGBoxed_Type, "GBoxed", G_TYPE_BOXED);
     REGISTER_GTYPE(d, PyGPointer_Type, "GPointer", G_TYPE_POINTER); 
@@ -2457,7 +2794,14 @@ initgobject(void)
 
     REGISTER_TYPE(d, PyGMainLoop_Type, "MainLoop"); 
     REGISTER_TYPE(d, PyGMainContext_Type, "MainContext"); 
-    
+
+    REGISTER_TYPE(d, PyGIOChannel_Type, "IOChannel");
+
+    REGISTER_TYPE(d, PyGSource_Type, "Source");
+    REGISTER_TYPE(d, PyGIdle_Type, "Idle");
+    REGISTER_TYPE(d, PyGTimeout_Type, "Timeout");
+    REGISTER_TYPE(d, PyGPollFD_Type, "PollFD");
+
     /* glib version */
     tuple = Py_BuildValue ("(iii)", glib_major_version, glib_minor_version,
 			   glib_micro_version);
@@ -2504,6 +2848,24 @@ initgobject(void)
     PyModule_AddIntConstant(m, "IO_HUP",  G_IO_HUP);
     PyModule_AddIntConstant(m, "IO_NVAL", G_IO_NVAL);
 
+#define addint(x) PyModule_AddIntConstant(m, #x, G_##x)
+
+    addint(IO_STATUS_ERROR);
+    addint(IO_STATUS_NORMAL);
+    addint(IO_STATUS_EOF);
+    addint(IO_STATUS_AGAIN);
+
+    addint(IO_FLAG_APPEND);
+    addint(IO_FLAG_NONBLOCK);
+    addint(IO_FLAG_IS_READABLE);
+    addint(IO_FLAG_IS_WRITEABLE);
+    addint(IO_FLAG_IS_SEEKABLE);
+    addint(IO_FLAG_MASK);
+    addint(IO_FLAG_GET_MASK);
+    addint(IO_FLAG_SET_MASK);
+
+#undef addint
+
     PyModule_AddIntConstant(m, "SPAWN_LEAVE_DESCRIPTORS_OPEN", G_SPAWN_LEAVE_DESCRIPTORS_OPEN);
     PyModule_AddIntConstant(m, "SPAWN_DO_NOT_REAP_CHILD", G_SPAWN_DO_NOT_REAP_CHILD);
     PyModule_AddIntConstant(m, "SPAWN_SEARCH_PATH", G_SPAWN_SEARCH_PATH);
@@ -2537,7 +2899,7 @@ initgobject(void)
     PyModule_AddObject(m, "TYPE_OBJECT", pyg_type_wrapper_new(G_TYPE_OBJECT));
     PyModule_AddObject(m, "TYPE_PYOBJECT", pyg_type_wrapper_new(PY_TYPE_OBJECT));
 
-    pyg_register_boxed_custom(G_TYPE_STRV,
+    pyg_register_gtype_custom(G_TYPE_STRV,
 			      _pyg_strv_from_gvalue,
 			      _pyg_strv_to_gvalue);
 }
