@@ -21,11 +21,6 @@
  */
 #include "pygobject-private.h"
 
-static const gchar *pygobject_class_id     = "PyGObject::class";
-static GQuark       pygobject_class_key    = 0;
-static const gchar *pygobject_wrapper_id   = "PyGObject::wrapper";
-static GQuark       pygobject_wrapper_key  = 0;
-
 static void pygobject_dealloc(PyGObject *self);
 static int  pygobject_traverse(PyGObject *self, visitproc visit, void *arg);
 static int  pygobject_clear(PyGObject *self);
@@ -72,6 +67,11 @@ pygobject_register_sinkfunc(GType type, void (* sinkfunc)(GObject *object))
 {
     SinkFunc sf;
 
+#if 0
+    g_return_if_fail(G_TYPE_IS_OBJECT(type));
+#endif
+    g_return_if_fail(sinkfunc != NULL);
+    
     if (!sink_funcs)
 	sink_funcs = g_array_new(FALSE, FALSE, sizeof(SinkFunc));
 
@@ -80,6 +80,372 @@ pygobject_register_sinkfunc(GType type, void (* sinkfunc)(GObject *object))
     g_array_append_val(sink_funcs, sf);
 }
 
+typedef struct {
+    PyObject_HEAD
+    GParamSpec **props;
+    guint n_props;
+    guint index;
+} PyGPropsIter;
+
+static void
+pyg_props_iter_dealloc(PyGPropsIter *self)
+{
+    g_free(self->props);
+    PyObject_Del((PyObject*) self);
+}
+
+static PyObject*
+pygobject_props_iter_next(PyGPropsIter *iter)
+{
+    if (iter->index < iter->n_props)
+        return pyg_param_spec_new(iter->props[iter->index++]);
+    else {
+        PyErr_SetNone(PyExc_StopIteration);
+        return NULL;
+    }
+}
+
+
+PyTypeObject PyGPropsIter_Type = {
+	PyObject_HEAD_INIT(NULL)
+	0,					/* ob_size */
+	"gobject.GPropsIter",			/* tp_name */
+	sizeof(PyGPropsIter),			/* tp_basicsize */
+	0,					/* tp_itemsize */
+	(destructor)pyg_props_iter_dealloc,	/* tp_dealloc */
+	0,					/* tp_print */
+	0,					/* tp_getattr */
+	0,					/* tp_setattr */
+	0,					/* tp_compare */
+	0,					/* tp_repr */
+	0,					/* tp_as_number */
+	0,					/* tp_as_sequence */
+	0,		       			/* tp_as_mapping */
+	0,					/* tp_hash */
+	0,					/* tp_call */
+	0,					/* tp_str */
+	0,					/* tp_getattro */
+	0,					/* tp_setattro */
+	0,					/* tp_as_buffer */
+	Py_TPFLAGS_DEFAULT,			/* tp_flags */
+ 	"GObject properties iterator",		/* tp_doc */
+	0,					/* tp_traverse */
+ 	0,					/* tp_clear */
+	0,					/* tp_richcompare */
+	0,					/* tp_weaklistoffset */
+	0,					/* tp_iter */
+	(iternextfunc)pygobject_props_iter_next, /* tp_iternext */
+};
+
+typedef struct {
+    PyObject_HEAD
+    /* a reference to the object containing the properties */
+    PyGObject *pygobject;
+    GType      gtype;
+} PyGProps;
+
+static void
+PyGProps_dealloc(PyGProps* self)
+{
+    PyGObject *tmp;
+
+    PyObject_GC_UnTrack((PyObject*)self);
+
+    tmp = self->pygobject;
+    self->pygobject = NULL;
+    Py_XDECREF(tmp);
+
+    PyObject_GC_Del((PyObject*)self);
+}
+
+static PyObject*
+build_parameter_list(GObjectClass *class)
+{
+    GParamSpec **props;
+    guint n_props = 0, i;
+    PyObject *prop_str;
+    char *name;
+    PyObject *props_list;
+
+    props = g_object_class_list_properties(class, &n_props);
+    props_list = PyList_New(n_props);
+    for (i = 0; i < n_props; i++) {
+	name = g_strdup(g_param_spec_get_name(props[i]));
+	/* hyphens cannot belong in identifiers */
+	g_strdelimit(name, "-", '_');
+	prop_str = PyString_FromString(name);
+	
+	PyList_SetItem(props_list, i, prop_str);
+    }
+
+    if (props)
+        g_free(props);
+    
+    return props_list;
+}
+
+static PyObject*
+PyGProps_getattro(PyGProps *self, PyObject *attr)
+{
+    char *attr_name;
+    GObjectClass *class;
+    GParamSpec *pspec;
+    GValue value = { 0, };
+    PyObject *ret;
+
+    attr_name = PyString_AsString(attr);
+    if (!attr_name) {
+        PyErr_Clear();
+        return PyObject_GenericGetAttr((PyObject *)self, attr);
+    }
+
+    class = g_type_class_ref(self->gtype);
+    
+    if (!strcmp(attr_name, "__members__")) {
+	return build_parameter_list(class);
+    }
+
+    pspec = g_object_class_find_property(class, attr_name);
+    g_type_class_unref(class);
+
+    if (!pspec) {
+	return PyObject_GenericGetAttr((PyObject *)self, attr);
+    }
+
+    if (!(pspec->flags & G_PARAM_READABLE)) {
+	PyErr_Format(PyExc_TypeError,
+		     "property '%s' is not readable", attr_name);
+	return NULL;
+    }
+
+    /* If we're doing it without an instance, return a GParamSpec */
+    if (!self->pygobject) {
+        return pyg_param_spec_new(pspec);
+    }
+    
+    g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(pspec));
+    g_object_get_property(self->pygobject->obj, attr_name, &value);
+    ret = pyg_param_gvalue_as_pyobject(&value, TRUE, pspec);
+    g_value_unset(&value);
+    
+    return ret;
+}
+
+static gboolean
+set_property_from_pspec(GObject *obj,
+			char *attr_name,
+			GParamSpec *pspec,
+			PyObject *pvalue)
+{
+    GValue value = { 0, };
+
+    if (pspec->flags & G_PARAM_CONSTRUCT_ONLY) {
+	PyErr_Format(PyExc_TypeError,
+		     "property '%s' can only be set in constructor",
+		     attr_name);
+	return FALSE;
+    }	
+
+    if (!(pspec->flags & G_PARAM_WRITABLE)) {
+	PyErr_Format(PyExc_TypeError,
+		     "property '%s' is not writable", attr_name);
+	return FALSE;
+    }	
+
+    g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(pspec));
+    if (pyg_param_gvalue_from_pyobject(&value, pvalue, pspec) < 0) {
+	PyErr_SetString(PyExc_TypeError,
+			"could not convert argument to correct param type");
+	return FALSE;
+    }
+
+    g_object_set_property(obj, attr_name, &value);
+    g_value_unset(&value);
+    
+    return TRUE;
+}
+
+static int
+PyGProps_setattro(PyGProps *self, PyObject *attr, PyObject *pvalue)
+{
+    GParamSpec *pspec;
+    char *attr_name;
+    GObject *obj;
+    
+    if (pvalue == NULL) {
+	PyErr_SetString(PyExc_TypeError, "properties cannot be "
+			"deleted");
+	return -1;
+    }
+
+    attr_name = PyString_AsString(attr);
+    if (!attr_name) {
+        PyErr_Clear();
+        return PyObject_GenericSetAttr((PyObject *)self, attr, pvalue);
+    }
+
+    if (!self->pygobject) {
+        PyErr_SetString(PyExc_TypeError,
+			"cannot set GOject properties without an instance");
+        return -1;
+    }
+
+    obj = self->pygobject->obj;
+    pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(obj), attr_name);
+    if (!pspec) {
+	return PyObject_GenericSetAttr((PyObject *)self, attr, pvalue);
+    }
+
+    if (!set_property_from_pspec(obj, attr_name, pspec, pvalue))
+	return -1;
+				  
+    return 0;
+}
+
+static int
+pygobject_props_traverse(PyGProps *self, visitproc visit, void *arg)
+{
+    if (self->pygobject && visit((PyObject *) self->pygobject, arg) < 0)
+        return -1;
+    return 0;
+}
+
+static PyObject*
+pygobject_props_get_iter(PyGProps *self)
+{
+    PyGPropsIter *iter;
+    GObjectClass *class;
+
+    iter = PyObject_NEW(PyGPropsIter, &PyGPropsIter_Type);
+    class = g_type_class_ref(self->gtype);
+    iter->props = g_object_class_list_properties(class, &iter->n_props);
+    iter->index = 0;
+    g_type_class_unref(class);
+    return (PyObject *) iter;
+}
+
+static int
+PyGProps_length(PyGProps *self)
+{
+    GObjectClass *class;
+    int n_props;
+    
+    class = g_type_class_ref(self->gtype);
+    g_object_class_list_properties(class, &n_props);
+    g_type_class_unref(class);
+    
+    return n_props;
+}
+
+static PySequenceMethods _PyGProps_as_sequence = {
+    (inquiry)PyGProps_length,
+    (binaryfunc)0,
+    (intargfunc)0,
+    (intargfunc)0,
+    (intintargfunc)0,
+    (intobjargproc)0,
+    (intintobjargproc)0
+};
+
+PyTypeObject PyGProps_Type = {
+    PyObject_HEAD_INIT(NULL)
+    0,                                          /* ob_size */
+    "gobject.GProps",                           /* tp_name */
+    sizeof(PyGProps),                           /* tp_basicsize */
+    0,                                          /* tp_itemsize */
+    (destructor)PyGProps_dealloc,               /* tp_dealloc */
+    0,                                          /* tp_print */
+    0,                                          /* tp_getattr */
+    0,                                          /* tp_setattr */
+    0,                                          /* tp_compare */
+    0,                                          /* tp_repr */
+    0,                                          /* tp_as_number */
+    (PySequenceMethods*)&_PyGProps_as_sequence, /* tp_as_sequence */
+    0,                                          /* tp_as_mapping */
+    0,                                          /* tp_hash */
+    0,                                          /* tp_call */
+    0,                                          /* tp_str */
+    (getattrofunc)PyGProps_getattro,            /* tp_getattro */
+    (setattrofunc)PyGProps_setattro,            /* tp_setattro */
+    0,                                          /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC,      /* tp_flags */
+    "The properties of the GObject accessible as "
+    "Python attributes.",                       /* tp_doc */
+    (traverseproc)pygobject_props_traverse,     /* tp_traverse */
+    0,                                          /* tp_clear */
+    0,                                          /* tp_richcompare */
+    0,                                          /* tp_weaklistoffset */
+    (getiterfunc)pygobject_props_get_iter,      /* tp_iter */
+    0,                                          /* tp_iternext */
+};
+
+
+static PyObject *
+pyg_props_descr_descr_get(PyObject *self, PyObject *obj, PyObject *type)
+{
+    PyGProps *gprops;
+
+    gprops = PyObject_GC_New(PyGProps, &PyGProps_Type);
+    if (obj == NULL || obj == Py_None) {
+        gprops->pygobject = NULL;
+        gprops->gtype = pyg_type_from_object(type);
+    } else {
+        if (!PyObject_IsInstance(obj, (PyObject *) &PyGObject_Type)) {
+            PyErr_SetString(PyExc_TypeError, "cannot use GObject property"
+                            " descriptor on non-GObject instances");
+            return NULL;
+        }
+        Py_INCREF(obj);
+        gprops->pygobject = (PyGObject *) obj;
+        gprops->gtype = pyg_type_from_object(obj);
+    }
+    return (PyObject *) gprops;
+}
+
+PyTypeObject PyGPropsDescr_Type = {
+	PyObject_HEAD_INIT(NULL)
+	0,					/* ob_size */
+	"gobject.GPropsDescr",			/* tp_name */
+	0,					/* tp_basicsize */
+	0,					/* tp_itemsize */
+	0,	 				/* tp_dealloc */
+	0,					/* tp_print */
+	0,					/* tp_getattr */
+	0,					/* tp_setattr */
+	0,					/* tp_compare */
+	0,					/* tp_repr */
+	0,					/* tp_as_number */
+	0,					/* tp_as_sequence */
+	0,		       			/* tp_as_mapping */
+	0,					/* tp_hash */
+	0,					/* tp_call */
+	0,					/* tp_str */
+	0,					/* tp_getattro */
+	0,					/* tp_setattro */
+	0,					/* tp_as_buffer */
+	Py_TPFLAGS_DEFAULT,			/* tp_flags */
+ 	0,					/* tp_doc */
+	0,					/* tp_traverse */
+ 	0,					/* tp_clear */
+	0,					/* tp_richcompare */
+	0,					/* tp_weaklistoffset */
+	0,					/* tp_iter */
+	0,					/* tp_iternext */
+	0,					/* tp_methods */
+	0,					/* tp_members */
+	0,					/* tp_getset */
+	0,					/* tp_base */
+	0,					/* tp_dict */
+	pyg_props_descr_descr_get,		/* tp_descr_get */
+	0,					/* tp_descr_set */
+	0,					/* tp_dictoffset */
+	0,					/* tp_init */
+	0,					/* tp_alloc */
+	0,					/* tp_new */
+	0,               			/* tp_free */
+};
+
+
 /**
  * pygobject_register_class:
  * @dict: the module dictionary.  A reference to the type will be stored here.
@@ -101,15 +467,14 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
     PyObject *o;
     const char *class_name, *s;
 
-    if (!pygobject_class_key)
-	pygobject_class_key = g_quark_from_static_string(pygobject_class_id);
 
     class_name = type->tp_name;
     s = strrchr(class_name, '.');
     if (s != NULL)
 	class_name = s + 1;
 	
-    type->ob_type = &PyType_Type;
+    type->ob_type = &PyGObject_MetaType;
+
     if (bases) {
 	type->tp_bases = bases;
 	type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
@@ -152,9 +517,6 @@ pygobject_register_wrapper(PyObject *self)
 {
     GObject *obj = ((PyGObject *)self)->obj;
 
-    if (!pygobject_wrapper_key)
-	pygobject_wrapper_key=g_quark_from_static_string(pygobject_wrapper_id);
-
     sink_object(obj);
     Py_INCREF(self);
     g_object_set_qdata_full(obj, pygobject_wrapper_key, self,
@@ -169,7 +531,7 @@ pygobject_register_wrapper(PyObject *self)
  * bases. It will currently not filter out interfaces already implemented by
  * it parents.
  *
- * Returns: a PyTypeObject for the new type of NULL if it couldn't be created
+ * Returns: a PyTypeObject for the new type or NULL if it couldn't be created
  */
 PyTypeObject *
 pygobject_new_with_interfaces(GType gtype)
@@ -277,9 +639,6 @@ pygobject_new_with_interfaces(GType gtype)
 	    PyDict_SetItemString(mod_dict, gtype_name, (PyObject *)type);
     }
 
-    if (!pygobject_class_key)
-	pygobject_class_key = g_quark_from_static_string(pygobject_class_id);
-
     /* stash a pointer to the python class with the GType */
     Py_INCREF(type);
     g_type_set_qdata(gtype, pygobject_class_key, type);
@@ -322,6 +681,7 @@ pygobject_lookup_class(GType gtype)
 /**
  * pygobject_new:
  * @obj: a GObject instance.
+ * @sink: whether to sink any floating reference found on the GObject.
  *
  * This function gets a reference to a wrapper for the given GObject
  * instance.  If a wrapper has already been created, a new reference
@@ -331,18 +691,15 @@ pygobject_lookup_class(GType gtype)
  * Returns: a reference to the wrapper for the GObject.
  */
 PyObject *
-pygobject_new(GObject *obj)
+pygobject_new_full(GObject *obj, gboolean sink)
 {
     PyGObject *self;
 
-    if (!pygobject_wrapper_key)
-	pygobject_wrapper_key = g_quark_from_static_string(pygobject_wrapper_id);
-
     if (obj == NULL) {
 	Py_INCREF(Py_None);
 	return Py_None;
     }
-    
+
     /* we already have a wrapper for this object -- return it. */
     self = (PyGObject *)g_object_get_qdata(obj, pygobject_wrapper_key);
     if (self != NULL) {
@@ -360,7 +717,8 @@ pygobject_new(GObject *obj)
         pyg_begin_allow_threads;
 	self->obj = g_object_ref(obj);
 	pyg_end_allow_threads;
-	sink_object(self->obj);
+        if (sink)
+            sink_object(self->obj);
 
 	self->inst_dict = NULL;
 	self->weakreflist = NULL;
@@ -376,6 +734,12 @@ pygobject_new(GObject *obj)
     return (PyObject *)self;
 }
 
+PyObject *
+pygobject_new(GObject *obj)
+{
+    return pygobject_new_full(obj, TRUE);
+}
+
 static void
 pygobject_unwatch_closure(gpointer data, GClosure *closure)
 {
@@ -608,12 +972,8 @@ pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
 	    n_params++;
 	}
     }
-
-    self->obj = g_object_newv(object_type, n_params, params);
-    if (self->obj)
-	pygobject_register_wrapper((PyObject *)self);
-    else
-	PyErr_SetString (PyExc_RuntimeError, "could not create object");
+    if (pygobject_constructv(self, n_params, params))
+	PyErr_SetString(PyExc_RuntimeError, "could not create object");
 	   
  cleanup:
     for (i = 0; i < n_params; i++) {
@@ -669,7 +1029,6 @@ pygobject_set_property(PyGObject *self, PyObject *args)
 {
     gchar *param_name;
     GParamSpec *pspec;
-    GValue value = { 0, };
     PyObject *pvalue;
 
     if (!PyArg_ParseTuple(args, "sO:GObject.set_property", &param_name,
@@ -682,19 +1041,10 @@ pygobject_set_property(PyGObject *self, PyObject *args)
 			"the object does not support the given parameter");
 	return NULL;
     }
-    if (!(pspec->flags & G_PARAM_WRITABLE)) {
-	PyErr_Format(PyExc_TypeError, "property %s is not writable",
-		     param_name);
+    
+    if (!set_property_from_pspec(self->obj, param_name, pspec, pvalue))
 	return NULL;
-    }
-    g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(pspec));
-    if (pyg_param_gvalue_from_pyobject(&value, pvalue, pspec) < 0) {
-	PyErr_SetString(PyExc_TypeError,
-			"could not convert argument to correct param type");
-	return NULL;
-    }
-    g_object_set_property(self->obj, param_name, &value);
-    g_value_unset(&value);
+    
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -754,7 +1104,7 @@ pygobject_set_data(PyGObject *self, PyObject *args)
     GQuark quark;
     PyObject *data;
 
-    if (!PyArg_ParseTuple(args, "sO:GObject.get_data", &key, &data))
+    if (!PyArg_ParseTuple(args, "sO:GObject.set_data", &key, &data))
 	return NULL;
     quark = g_quark_from_string(key);
     Py_INCREF(data);
@@ -1253,3 +1603,91 @@ PyTypeObject PyGObject_Type = {
     (PyObject *)0,			/* tp_bases */
 };
 
+
+static int
+pygobjectmeta_register(PyTypeObject *subtype, PyObject *instance_dict)
+{
+    PyObject *pytype_name;
+    char *type_name = NULL;
+    int retval = 0;
+    
+    if (!instance_dict) {
+        PyErr_Clear();
+	goto out;
+    }
+
+    /* If it's already registered, skip registration */
+    if (PyDict_GetItemString(instance_dict, "__gtype__"))
+	goto out;
+
+    pytype_name = PyDict_GetItemString(instance_dict, "__gtype_name__");
+    if (pytype_name)
+	type_name = g_strdup(PyString_AsString(pytype_name));
+    retval = pyg_type_register(subtype, type_name);
+
+out:
+    return retval;
+}
+
+/* -------- GObject MetaClass -----------*/
+static int
+pygobjectmeta_init(PyTypeObject *subtype, PyObject *args, PyObject *kwargs)
+{
+    PyObject *instance_dict;
+    
+    if (PyType_Type.tp_init((PyObject *) subtype, args, kwargs))
+        return -1;
+    
+    instance_dict = PyTuple_GetItem(args, 2);
+    return pygobjectmeta_register(subtype, instance_dict);
+}
+
+
+PyTypeObject PyGObject_MetaType = {
+    PyObject_HEAD_INIT(NULL)
+    0,					/* ob_size */
+    "gobject.GObjectMeta",		/* tp_name */
+    0,					/* tp_basicsize */
+    0,					/* tp_itemsize */
+    /* methods */
+    (destructor)0,			/* tp_dealloc */
+    (printfunc)0,			/* tp_print */
+    (getattrfunc)0,			/* tp_getattr */
+    (setattrfunc)0,			/* tp_setattr */
+    (cmpfunc)0,				/* tp_compare */
+    (reprfunc)0,			/* tp_repr */
+    0,					/* tp_as_number */
+    0,					/* tp_as_sequence */
+    0,					/* tp_as_mapping */
+    (hashfunc)0,			/* tp_hash */
+    (ternaryfunc)0,			/* tp_call */
+    (reprfunc)0,			/* tp_str */
+    (getattrofunc)0,			/* tp_getattro */
+    (setattrofunc)0,			/* tp_setattro */
+    0,					/* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_BASETYPE, /* tp_flags */
+    "GObject Metaclass -- automatically registers GObject subclasses "
+    "as new GType's", /* Documentation string */
+    (traverseproc)0,			/* tp_traverse */
+    (inquiry)0,				/* tp_clear */
+    (richcmpfunc)0,			/* tp_richcompare */
+    0,					/* tp_weaklistoffset */
+    (getiterfunc)0,			/* tp_iter */
+    (iternextfunc)0,			/* tp_iternext */
+    0,                                  /* tp_methods */
+    0,					/* tp_members */
+    0,					/* tp_getset */
+    &PyType_Type,			/* tp_base */
+    (PyObject *)0,			/* tp_dict */
+    0,					/* tp_descr_get */
+    0,					/* tp_descr_set */
+    0,					/* tp_dictoffset */
+    (initproc)pygobjectmeta_init,	/* tp_init */
+    (allocfunc)0,			/* tp_alloc */
+    (newfunc)0,				/* tp_new */
+    0,					/* tp_free */
+    (inquiry)0,				/* tp_is_gc */
+    NULL,				/* tp_bases */
+};
+
+
